/**
 * @file aufgabe01.cpp
 * @author Gießel, Franjo (Franjo.Giessel@smail.HS-Bremerhaven.de)
 * @brief Implementieren Sie einen binären Suchbaum über unsigned int. Denken Sie an den Destruktor, Copykonstruktor, Zuweisungsoperator und Ausgabeoperator.  
 * @version 0.1
 * @date 2022-07-08
 * 
 * @copyright Copyright (c) 2022
 * 
 */

#include <iostream>
#include <assert.h>
using namespace std;

/**
 * @brief Representation of a node in a binary tree
 * 
 */
class Node {
    public:
        /**
         * @brief Construct a new Node object
         * 
         * @param key Key-Value
         * @param data Data-Value
         */
        Node (unsigned key, unsigned data) :
        mu_key(key),
        mu_data(data)
        {

        }

        /**
         * @brief Copys a Node object
         * 
         * @param crArg 
         */
        Node (const Node& crArg) :
        mu_key(crArg.mu_key),
        mu_data(crArg.mu_data),
        mo_left(crArg.mo_left != nullptr ? new Node (*crArg.mo_left) : nullptr),
        mo_right(crArg.mo_right != nullptr ? new Node (*crArg.mo_right) : nullptr)
        {
        }

        /**
         * @brief Destroy the Node object
         * 
         */
        ~Node () {
            // Delete auf Nullptr gibt keinen Fehler
            delete mo_left;
            delete mo_right;
        }

        /**
         * @brief Get the Key value
         * 
         * @return unsigned Key-value
         */
        unsigned getKey () const {
            return mu_key;
        }

        /**
         * @brief Get the next Left-node object
         * 
         * @return Node
         */
        Node* getLeft () const {
            return mo_left;
        }

        /**
         * @brief Set the next Left-node object
         * 
         * @param left 
         */
        void setLeft (Node* left) {
            mo_left = left;
        }

        /**
         * @brief Get the next Right-node object
         * 
         * @return Node
         */
        Node* getRight () const {
            return mo_right;
        }

        /**
         * @brief Set the next Right-node object
         * 
         * @param right 
         */
        void setRight (Node* right) {
            mo_right = right;
        }

        Node& operator= (const Node crArg) = delete;

        friend ostream& operator<< (ostream&, const Node& crArg);

    
    private:
        /**
         * @brief Prints the space infront of a node-entry.
         * TODO: correct the lines
         * 
         * @param os Output stream
         * @param versatz Number of whitespaces
         */
        void printVersatz (ostream& os, unsigned versatz) const {
            for (unsigned i = 0; i < versatz; ++i) {
                os << (i % 2 == 0 ? "│" : " ");
            }
        }

        /**
         * @brief Prints the node and its childs to os
         * 
         * @param os Output stream
         * @param versatz Number of whitespaces
         * @param ebene Not in use
         */
        void printNode (ostream& os, unsigned versatz, unsigned ebene) const {
            os << mu_data << endl;
            printVersatz(os, versatz);
            os << "├─";
            if (mo_left != nullptr) {
                mo_left->printNode(os, versatz+2, ebene+1);
            }
            os << endl;
            printVersatz(os, versatz);
            os << "└─";
            if (mo_right != nullptr) {
                mo_right->printNode(os, versatz+2, ebene+1);
            }
        }
        
        const unsigned mu_key;
        const unsigned mu_data;
        Node* mo_left = nullptr;
        Node* mo_right = nullptr;
};

/**
 * @brief The Ausgabeoperator prints the node and its childs to os
 * 
 * @param os Outputstream
 * @param crArg Node that should be printed
 * @return ostream& Outputstream
 */
ostream& operator<< (ostream& os, const Node& crArg) {
    crArg.printNode(os, 0, 0);
    return os;
}

/**
 * @brief Representation of a binary tree
 * 
 */
class BinTree {
    public:
        BinTree () = default;

        /**
         * @brief Copys a Node object
         * 
         * @param crArg 
         */
        BinTree (const BinTree& crArg) :
        mo_root(crArg.mo_root != nullptr ? new Node (*crArg.mo_root) : nullptr) 
        {

        }

        /**
         * @brief Destroy the Bin Tree object
         * 
         */
        ~BinTree () {
            delete mo_root;
        }
    
    private:
        /**
         * @brief Internal NodeHandler-class used for the remove-methode of BinTree
         * 
         */
        class NodeHandler {
            public:
                /**
                 * @brief Construct a new Node Handler object
                 * 
                 * @param n Current node
                 * @param root Root node of the binary tree
                 */
                NodeHandler (Node* n, Node* root) :
                mo_node(n),
                mo_root(root)
                {
                }

                /**
                 * @brief Step one node deeper. Which node to be used is specified in de left parameter
                 * 
                 * @param left Use the left node
                 */
                void down (bool left) {
                    mo_dad = mo_node;
                    mo_node = left ? mo_node->getLeft() : mo_node->getRight();
                }

                /**
                 * @brief Returns, if the current node is nullptr
                 * 
                 * @return true Current node is nullptr
                 * @return false Current node is not nullptr
                 */
                bool isNull () const {
                    return mo_node == nullptr;
                }

                /**
                 * @brief Get the current Node object
                 * 
                 * @return Node* current Node object
                 */
                Node* getNode () const {
                    return mo_node;
                }

                /**
                 * @brief Get the Dad-node object
                 * 
                 * @return Node* dad-node object
                 */
                Node* getDad () const {
                    return mo_dad;
                }

                /**
                 * @brief Get the Key of the current node object
                 * 
                 * @return unsigned Key of the current node object
                 */
                unsigned getKey () const {
                    return mo_node->getKey();
                }

                /**
                 * @brief Set the current node object to the parameter
                 * 
                 * @param n New node object
                 */
                void set (Node* n) {
                    assert(mo_node != nullptr);
                    if (mo_dad == nullptr) {
                        mo_root = n;
                    } else if (mo_node != nullptr ? mo_node == mo_dad->getLeft() : n->getKey() < mo_dad->getKey()){
                        mo_dad->setLeft(n);
                    } else {
                        mo_dad->setRight(n);
                    }
                    mo_node = n;
                }

            private:
                Node* mo_dad = nullptr;
                Node* mo_node = nullptr;
                Node* mo_root = nullptr;
        };

    public:

        /**
         * @brief Searches in the binary tree for the node with the key parameter.
         * If a node is found, the pointer to the node is returned.
         * If no node is found, a nullpointer will be returned.
         * 
         * @param key Key to be searched
         * @return Node* Pointer to the node object. Is null, if object can not be found
         */
        Node* search (unsigned key) const {
            Node* tmp = mo_root;
            while (tmp != nullptr) {
                if (key < tmp->getKey()) {
                    tmp = tmp->getLeft();
                } else if (tmp->getKey() < key) {
                    tmp = tmp->getRight();
                } else {
                    return tmp;
                }
            }
            return nullptr;
        }

        /**
         * @brief Inserts a new node object in the binary tree
         * 
         * @param key Key-value of the new node
         * @param data Data-value of the new node
         */
        void insert (unsigned key, unsigned data) {
            mo_root = insert (mo_root, key, data);
        }

        /**
         * @brief Removes the node with the key.
         * 
         * @param key Key of the node object
         * @return true Object was deleted
         * @return false Object was not found and not deleted
         */
        bool remove (unsigned key) {
            NodeHandler h(mo_root, mo_root);
            while (!h.isNull()) {
                if (key == h.getKey()) {
                    if (h.getNode()->getRight() == nullptr) {
                        h.set(h.getNode()->getLeft());
                    } else {
                        NodeHandler h2(h.getNode(), mo_root);
                        h2.down(false);
                        while (!h2.isNull()) {
                            h2.down(true);
                        }
                        h2.set(h.getNode()->getLeft());
                        h.set(h.getNode()->getRight());
                    }
                    return true;
                }
                h.down(key < h.getKey());
            }
            return false;
        }

        /**
         * @brief Get the Root node object of the binary tree
         * 
         * @return Node* Root node object
         */
        Node* getRoot () const {
            return mo_root;
        }

        BinTree& operator= (BinTree crArg) {
            Node* tmp = crArg.mo_root;
            crArg.mo_root = mo_root;
            mo_root = tmp;
            return *this;
        }

        friend ostream& operator<< (ostream&, const Node& crArg);

    private:
        /**
         * @brief Inserts a new node object in the binary tree
         * 
         * @param n Start node
         * @param key Key-value of the new node
         * @param data Data-value of the new node
         * @return Node* Returns the new inserted node
         */
        Node* insert (Node* n, unsigned key, unsigned data) {
            if (n == nullptr) { // kann durch if (n) ersetzt werden, weil nullptr == false?
                return new Node(key, data);
            } else {
                if (key < n->getKey()) {
                    n->setLeft(insert(n->getLeft(), key, data));
                } else if (key > n->getKey()) {
                    n->setRight(insert(n->getRight(), key, data));
                }
                return n;
            }
        }

        Node* mo_root = nullptr;

};

ostream& operator<< (ostream& os, const BinTree& crArg) {
    os << *crArg.getRoot();
    return os;
}

int main () {
    BinTree bt;
    bt.insert(34, 34);
    bt.insert(13, 13);
    bt.insert(40, 40);
    bt.insert(5, 5);
    // bt.insert(17, 17);
    // bt.insert(38, 38);
    // bt.insert(42, 42);
    // bt.insert(15, 15);
    // bt.insert(16, 16);
    // cout << bt << endl << endl;
    
    BinTree bt2 = bt;
    cout << bt2 << endl << endl;
    bt2.insert(35, 35);
    cout << bt2 << endl << endl;
    cout << bt << endl << endl;
    bt = bt2;
    cout << bt2 << endl << endl;
    cout << bt << endl << endl;
    bt.insert(17, 17);
    cout << bt2 << endl << endl;
    cout << bt << endl << endl;
    bt2.insert(38, 38);
    cout << bt2 << endl << endl;
    cout << bt << endl << endl;

    return 0;
}

/*
34
├─13
│ ├─5
│ │ ├─
│ │ └─
│ └─17
│   ├─15
│   │ ├─
│   │ └─16
│   │   ├─
│   │   └─
│   └─
└─40
  ├─38
  │ ├─
  │ └─
  └─42
    ├─
    └─
*/